% Graph based transform presented in G. Shen et al,
% "Edge-adaptive transforms for efficient depth map coding,"
% 28th Picture Coding Symposium, PCS'10, Dec. 2010.
% (The experimental results in the paper is generated not by this code
%  but using the C implementation based on H.264/AVC JM code.)
% This version does not include the removal of isolated nodes in the graph.
%
% Authors: Woo-Shik Kim, Godwin Shen
% updated Apr. 5 2011
%

% Transform and RD cost comparison to DCT

clear all;
close all
clc;

framenum = 0;

QP = [24 28 32 36];
nQP = length(QP);
QStep = zeros(1,nQP);
for i = 1:nQP
    QStep(i) = QStep_Compute(QP(i));
end
InName = 'res_ballet_1024x768_depth_view_37.yuv';
cols = 1024;
rows = 768;
[X u v] = extract_single_yuv(InName, cols, rows, framenum);
X = X-128;

[Nx,Ny] = size(X);

[BW_big, T] = edge_double_grid_alternate_tree(X);

blk_size = 4;

Nr = size(X,1)/blk_size;
Nc = size(X,2)/blk_size;
NN = 0;

PSNR = zeros(3,nQP);
MSE = zeros(3,nQP);
ReqCoeffs = zeros(3,nQP);
RateCoeffs = zeros(3,nQP);
RDcost = zeros(3,nQP);

NN_G = zeros(1,nQP);
mode_info = zeros(1,Nr*Nc);

for i = 1:Nr
    for j = 1:Nc
        ind_row = blk_size*(i-1)+1:blk_size*i;
        ind_col = blk_size*(j-1)+1:blk_size*j;
        ind_row_2 = 2*min(ind_row)-1:2*max(ind_row);
        ind_col_2 = 2*min(ind_col)-1:2*max(ind_col);
        BWb = BW_big(ind_row_2, ind_col_2);
        X_ij = X(ind_row,ind_col);
        
        BWb(2:2:2*blk_size,2:2:2*blk_size)=0;
        
        BWb_T = BWb;
        
        % Check for vertical edges (in odd rows, even columns of BWb)
        vert_check = sum(sum(BWb(1:2:2*blk_size-1,2:2:2*blk_size)));

        % Check for horizontal edges (in even rows, odd columns of BWb)
        hor_check = sum(sum(BWb(2:2:2*blk_size,1:2:2*blk_size-1)));
        
%         if sum(BWb(:)) ~= 0 && (vert_check > 0) && (hor_check > 0)
        if sum(BWb(:)) ~= 0 
            t0 = clock;
            A_ij = AdjMat(BWb);
            nG = length(A_ij);
            Xvec_ij = X_ij(:);
            % Spectral Transform and DCT Transform Computation
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            D_ij = sum(A_ij,2);
            D_ij = diag(D_ij);
            L_ij = D_ij - A_ij;
            [EV Lam] = eig(L_ij);
            Ybb = EV'*Xvec_ij; % Spectral Transform
            Yb_uq = dct2(X_ij); % DCT transform
            
            Yb_uq_ro = zigzag(Yb_uq);
            Yb_uq_ro = Yb_uq_ro(:);
                      
%             sum([Ybb(1:length(r)-1) Yb_uq_ro(1:length(r)-1)].^2)./sum([Ybb Yb_uq_ro].^2)
            
            [dummy, dummy2] = EncodeBlock_CABAC2(round(Ybb/QStep(1)),BWb,QP);
            rate_edges_G(1,1:nQP) = dummy2;       
            
            for ctrQP = 1:nQP
                lambda = 0.85*2^((QP(ctrQP)-12)./3);
                
                Yb_uq_hat = round(Yb_uq/QStep(ctrQP));
                Yb_uq_hat_rcvd = round(Yb_uq_hat*QStep(ctrQP))+0.5;
                Xhat_ij = idct2(Yb_uq_hat_rcvd);
                mse_dct = sum((Xhat_ij(:) - X_ij(:)).^2);
                MSE(1,ctrQP) = MSE(1,ctrQP) + mse_dct;
                ReqCoeffs(1,ctrQP) = ReqCoeffs(1,ctrQP)+ length(find(Yb_uq_hat(:)));

                Yb_ro = zigzag(Yb_uq_hat);
                rate_DCT(1,ctrQP) = blk_size.^2*entropy_mine(Yb_ro);
                RateCoeffs(1,ctrQP) = RateCoeffs(1,ctrQP) + rate_DCT(1,ctrQP);
                RDcost(1,ctrQP) = RDcost(1,ctrQP)+mse_dct + lambda.*rate_DCT(1,ctrQP);
                
                Ybb_hat = round(Ybb/QStep(ctrQP));
                Xvec_hat_ij = EV*(round(QStep(ctrQP)*Ybb_hat)+0.5);
                mse_lt = sum((Xvec_hat_ij -Xvec_ij).^2);
                
                rate_G(1,ctrQP) = blk_size.^2*entropy_mine(Ybb_hat);
                RDcost_G(1,ctrQP) = mse_lt + lambda.*(rate_G(1,ctrQP)+rate_edges_G(1,ctrQP));
                RDcost_DCT(1,ctrQP) = mse_dct + lambda.*rate_DCT(1,ctrQP);
                
                Xhat_dct{ctrQP}(ind_row,ind_col) = Xhat_ij;
                
                if RDcost_G(1,ctrQP) < (mse_dct + lambda.*rate_DCT(1,ctrQP))
                    MSE(2,ctrQP) = MSE(2,ctrQP) +  mse_lt;
                    ReqCoeffs(2,ctrQP) = ReqCoeffs(2,ctrQP)+ length(find(Ybb_hat));
                    RateCoeffs(2,ctrQP) = RateCoeffs(2,ctrQP) + rate_G(1,ctrQP) + rate_edges_G(1,ctrQP);
                    RDcost(2,ctrQP) = RDcost(2,ctrQP)+mse_lt + lambda.*(rate_G(1,ctrQP)+rate_edges_G(1,ctrQP));
                                        
                    mode_info(1,ctrQP) = 1;
                    NN_G(ctrQP) = NN_G(ctrQP)+1;
                    
                    disp(['QP = ', num2str(QP(ctrQP)), ', NN_G(QP) = ', num2str(NN_G(ctrQP))])
                    disp(['Block (', num2str(i), ', ', num2str(j), ')'])
                    
                    disp(['Rate(Total) : ', num2str(rate_G(1,ctrQP)+rate_edges_G(1,ctrQP))])
                    disp(['Rate(Coeff) : ', num2str(rate_G(1,ctrQP))])
                    disp(['Rate(Edges) : ', num2str(rate_edges_G(1,ctrQP))])
                    disp(['MSE(Graph) : ', num2str(mse_lt)])

                    disp(['Rate(DCT) : ', num2str(rate_DCT(1,ctrQP))])
                    disp(['MSE(DCT) : ', num2str(mse_dct)])

                    disp(['RDcost(Graph) : ', num2str(RDcost_G(1,ctrQP))])
                    disp(['RDcost(DCT) : ', num2str(RDcost_DCT(1,ctrQP))])
                    
                    1;
                else
                    MSE(2,ctrQP) = MSE(2,ctrQP) +  mse_dct;
                    ReqCoeffs(2,ctrQP) = ReqCoeffs(2,ctrQP)+ length(find(Yb_uq_hat));
                    RateCoeffs(2,ctrQP) = RateCoeffs(2,ctrQP) + rate_DCT(1,ctrQP);
                    RDcost(2,ctrQP) = RDcost(2,ctrQP)+mse_dct + lambda.*rate_DCT(1,ctrQP);
                end
                
            end
            NN = NN+1;
            disp(['Took ', num2str(etime(clock,t0)), ' s to compute metrics'])
        else
            
            Yb_uq = dct2(X_ij); % DCT transform
            for ctrQP = 1:nQP
                lambda = 0.85*2^((QP(ctrQP)-12)./3);

                Yb_uq_hat = round(Yb_uq/QStep(ctrQP));
                Yb_uq_hat_rcvd = round(Yb_uq_hat*QStep(ctrQP))+0.5;
                Xhat_ij = idct2(Yb_uq_hat_rcvd);
                mse_dct = sum((Xhat_ij(:) - X_ij(:)).^2);
                MSE(1,ctrQP) = MSE(1,ctrQP) + mse_dct;
                ReqCoeffs(1,ctrQP) = ReqCoeffs(1,ctrQP)+ length(find(Yb_uq_hat(:)));

                Yb_ro = zigzag(Yb_uq_hat);
                rate_DCT(1,ctrQP) = blk_size.^2*entropy_mine(Yb_ro);
                RateCoeffs(1,ctrQP) = RateCoeffs(1,ctrQP) + rate_DCT(1,ctrQP);
                RDcost(1,ctrQP) = RDcost(1,ctrQP)+mse_dct + lambda.*rate_DCT(1,ctrQP);
                
                MSE(2,ctrQP) = MSE(2,ctrQP) +  mse_dct;
                ReqCoeffs(2,ctrQP) = ReqCoeffs(2,ctrQP)+ length(find(Yb_uq_hat));
                RateCoeffs(2,ctrQP) = RateCoeffs(2,ctrQP) + rate_DCT(1,ctrQP);
                RDcost(2,ctrQP) = RDcost(2,ctrQP)+mse_dct + lambda.*rate_DCT(1,ctrQP);
            end
            NN = NN+1;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    end
end

MSE = MSE./(Nx*Ny);
ReqCoeffs = ReqCoeffs/(NN);
PSNR(1,:) = 10*log10(255^2 ./ MSE(1,:));
PSNR(2,:) = 10*log10(255^2 ./ MSE(2,:));

figure,
plot(RateCoeffs(1,:),PSNR(1,:),'r*-')
hold on
plot(RateCoeffs(2,:),PSNR(2,:),'o-')
legend('DCT transform','DCT / Spectral Transform');
title(['DCT vs Hybrid Transform (', num2str(blk_size), 'x', num2str(blk_size), ')'])
xlabel('bits')
ylabel('PSNR (dB)')
grid on
