% Graph based transform presented in G. Shen et al,
% "Edge-adaptive transforms for efficient depth map coding,"
% 28th Picture Coding Symposium, PCS'10, Dec. 2010.
% (The experimental results in the paper is generated not by this code
%  but using the C implementation based on H.264/AVC JM code.)
% This version does not include the removal of isolated nodes in the graph.
%
% Authors: Woo-Shik Kim, Godwin Shen
% Updated Apr. 17 2011
%

% Transform only without calculation of RD cost

clear all;
close all
clc;

framenum = 0;

%Residual signal
InName = 'res_ballet_1024x768_depth_view_37.yuv';
cols = 1024;
rows = 768;
blk_size = 4;
[X u v] = extract_single_yuv(InName, cols, rows, framenum);
X = X-128;
%Trnasform mode
% InName = 'disp2_transmodegbt_q24.y00';
% [transmode u v] = extract_single_yuv(InName, cols/blk_size, rows/blk_size, framenum);

[BW_big, T] = edge_double_grid_alternate_tree(X);

Nr = size(X,1)/blk_size;
Nc = size(X,2)/blk_size;

QP = 24;
QStep = QStep_Compute(QP);

for i = 1:Nr
    for j = 1:Nc
        ind_row = blk_size*(i-1)+1:blk_size*i;
        ind_col = blk_size*(j-1)+1:blk_size*j;
        ind_row_2 = 2*min(ind_row)-1:2*max(ind_row);
        ind_col_2 = 2*min(ind_col)-1:2*max(ind_col);
        BWb = BW_big(ind_row_2, ind_col_2);
        X_ij = X(ind_row,ind_col);
        
        BWb(2:2:2*blk_size,2:2:2*blk_size)=0;
        
        BWb_T = BWb;
        
%         % Check for vertical edges (in odd rows, even columns of BWb)
%         vert_check = sum(sum(BWb(1:2:2*blk_size-1,2:2:2*blk_size)));
% 
%         % Check for horizontal edges (in even rows, odd columns of BWb)
%         hor_check = sum(sum(BWb(2:2:2*blk_size,1:2:2*blk_size-1)));
%         
%         if sum(BWb(:)) ~= 0 && (vert_check > 0) && (hor_check > 0)
        if sum(BWb(:)) ~= 0
%         if transmode(i,j) == 1
            
            A_ij = AdjMat(BWb);
            nG = length(A_ij);
            Xvec_ij = X_ij(:);
            % Spectral Transform Computation
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            D_ij = sum(A_ij,2);
            D_ij = diag(D_ij);
            L_ij = D_ij - A_ij;
            [EV Lam] = eig(L_ij);
            Ybb = EV'*Xvec_ij; % Spectral Transform
            Ybb_hat = round(Ybb/QStep);
        else
            Ybb = dct2(X_ij); % DCT transform
            Ybb_hat = round(Ybb/QStep);
        end
    end
end
